// BACnetStackDLLClientExample.cpp : Defines the entry point for the console application.
// This application provides an example of how to implement a simple BACnet client device
// and how to register the various callbacks and hooks to the CAS BACnet Stack and how to
// send messages to BACnet server devices.
// For more information about the stack, please visit bacnetstack.com

#include "stdafx.h"

// Standard lib includes
#include <stdio.h>

// Adapter Includes
#include "CASBACnetStackAdapter.h"

// Resource and Common Includes
#include "SimpleUDP.h"
#include "CASBACnetStackExampleConstants.h"
#include "CLogging.h"
#include "common.h"
#include "endianness.h"

#include "BACnetPacket.h"

// Handle Keyboard input for linux
#ifdef __GNUC__
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>

int _kbhit(void) {
	struct termios oldt, newt;
	int ch;
	int oldf;

	tcgetattr(STDIN_FILENO, &oldt);
	newt = oldt;
	newt.c_lflag &= ~(ICANON | ECHO);
	tcsetattr(STDIN_FILENO, TCSANOW, &newt);
	oldf = fcntl(STDIN_FILENO, F_GETFL, 0);
	fcntl(STDIN_FILENO, F_SETFL, oldf | O_NONBLOCK);

	ch = getchar();

	tcsetattr(STDIN_FILENO, TCSANOW, &oldt);
	fcntl(STDIN_FILENO, F_SETFL, oldf);

	if (ch != EOF)
	{
		ungetc(ch, stdin);
		return 1;
	}

	return 0;
}
#endif //

// Application Defaults
#define DEFAULT_IP_ADDRESS "192.168.1.26"
#define DEFAULT_PORT 47808

// Constants
static const int BACNET_CLIENT_DEVICE_INSTANCE = 12345;
static const int MAX_XML_RENDER_BUFFER_LENGTH = 1024 * 10;

static const int BACNET_CLIENT_EXAMPLE_MAJOR_VERSION = 1;
static const int BACNET_CLIENT_EXAMPLE_MINOR_VERSION = 0;
static const int BACNET_CLIENT_EXAMPLE_PATCH_VERSION = 0;

// Globals
// =======================================
CSimpleUDP g_udp;	// UDP resource
char xmlRenderBuffer[MAX_XML_RENDER_BUFFER_LENGTH];
float analogValuePresentValue;

// Function Declarations
// =======================================
// Loading Functions
bool LoadDLLCallbackFunctions();					// Register Callback Functions to the DLL

// Setup Functions
bool SetupDLLFunctionsAndCallbacks();				// Loads the dll functions into function pointers and registers the callbacks
bool SetupBACnetDevice(uint32_t deviceInstance);	// Loads the client device

// Helper Functions
bool ParseCommandLine(int argc, char ** argv);
void WaitForResponse(unsigned int timeout = 3);

// Example Functions
void ExampleReadPropertyAnalogInputPresentValue();
void ExampleWhoIs();
void ExampleWhoHas();
void ExampleReadPropertyDevice();
void ExampleReadPropertyAnalogInput();
void ExampleReadPropertyAnalogValue();
void ExampleReadPropertyAnalogOutput();
void ExampleReadPropertyBinaryInput();
void ExampleReadPropertyBinaryOutput();
void ExampleReadPropertyBinaryValue();
void ExampleReadPropertyMultiStateInput();
void ExampleReadPropertyMultiStateOutput();
void ExampleReadPropertyMultiStateValue();
void ExampleReadPropertyBitstringValue();
void ExampleReadPropertyCharacterstringValue();
void ExampleReadPropertyDateValue();
void ExampleReadPropertyIntegerValue();
void ExampleReadPropertyLargeAnalogValue();
void ExampleReadPropertyOctetstringValue();
void ExampleReadPropertyPositiveIntegerValue();
void ExampleReadPropertyTimeValue();
void ExampleReadPropertyNetworkPort();
void ExampleWritePropertyAnalogValue();
void ExampleWritePropertyAnalogOutput();
void ExampleSubscribeCOV();
void ExampleCreateObjectByType();
void ExampleCreateObjectByTypeWithInitialValues();
void ExampleCreateObjectByIdentifier();
void ExampleCreateObjectByIdentifierWithInitialValues();
void ExampleDeleteObject();
void ExampleSendTimeSynchronization(); 
void ExampleSendUTCTimeSynchronization();
void ExampleReadRange();
void ExampleReadRangeByPosition();
void ExampleReadRangeBySequenceNumber();
void ExampleReadRangeByTime();
void ExampleReinitializeDeviceNoPassword();
void ExampleReinitializeDeviceWithPassword();
void ExampleDeviceCommunicationControlNoOptionalParameters();
void ExampleDeviceCommunicationControlWithOptionalParameters();
void ExampleReadPropertyAsync();
void ExampleWritePropertyAsync();

// Callback Functions to Register to the DLL
// Socket Functions
uint16_t CallbackReceiveMessage(uint8_t * message, const uint16_t maxMessageLength, uint8_t * receivedConnectionString, const uint8_t maxConnectionStringLength, uint8_t * receivedConnectionStringLength, uint8_t * networkType);
uint16_t CallbackSendMessage(const uint8_t * message, const uint16_t messageLength, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, bool broadcast);

// System Functions
time_t CallbackGetSystemTime();

// Hooks for unconfirmed request
void HookIAm(const uint32_t deviceIdentifier, const uint32_t maxApduLengthAccepted, const uint8_t segmentationSupported, const uint16_t vendorIdentifier, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength);
void HookIHave(const uint32_t deviceIdentifier, const uint16_t objectType, const uint32_t objectInstance, const char * objectName, const uint32_t objectNameLength, const uint8_t objectNameEncoding, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength);

// Hooks for simple PDUs
void HookError(const uint8_t originalInvokeId, const uint32_t errorChoice, const uint32_t errorClass, const uint32_t errorCode, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength, const bool useObjectProperty, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier);
void HookReject(const uint8_t originalInvokeId, const uint32_t rejectReason, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength);
void HookAbort(const uint8_t originalInvokeId, const bool sentByServer, const uint32_t abortReason, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength);
void HookSimpleAck(const uint8_t originalInvokeId, const uint32_t serverAckChoice, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength);
void HookTimeout(const uint8_t originalInvokeId, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength);

// Hooks for different ComplexAck PDUs
void HookPropertyBitString(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const bool * value, const uint32_t length, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength);
void HookPropertyBool(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const bool value, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength);
void HookPropertyCharString(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const char * value, const uint32_t length, const uint8_t encoding, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength);
void HookPropertyDate(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const uint8_t year, const uint8_t month, const uint8_t day, const uint8_t weekday, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength);
void HookPropertyDouble(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const double value, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength);
void HookPropertyEnum(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const uint32_t value, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength);
void HookPropertyNull(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength);
void HookPropertyObjectIdentifier(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const uint16_t objectTypeValue, const uint32_t objectInstanceValue, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength);
void HookPropertyOctString(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const uint8_t * value, const uint32_t length, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength);
void HookPropertyInt(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const int32_t value, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength);
void HookPropertyReal(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const float value, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength);
void HookPropertyTime(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const uint8_t hour, const uint8_t minute, const uint8_t second, const uint8_t hundrethSecond, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength);
void HookPropertyUInt(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const uint32_t value, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength);

std::string ipAddress;
uint16_t port;
uint8_t connectionString[6];
uint8_t invokeId;

int main(int argc, char ** argv)
{
	// Initialize defaults
	memset(xmlRenderBuffer, 0, MAX_XML_RENDER_BUFFER_LENGTH);
	ipAddress = std::string(DEFAULT_IP_ADDRESS);
	port = DEFAULT_PORT;
	invokeId = 0;

	// Handle command-line parameters
	if (argc > 1) {
		if (!ParseCommandLine(argc, argv)) {
			return -1;
		}
	}

	CLogging::Fyi("CAS BACnet Stack DLL Client Example v%d.%d.%d", BACNET_CLIENT_EXAMPLE_MAJOR_VERSION, BACNET_CLIENT_EXAMPLE_MINOR_VERSION, BACNET_CLIENT_EXAMPLE_PATCH_VERSION);
	CLogging::Fyi("================================================");
	CLogging::Fyi("");
	CLogging::Fyi("The CAS BACnet Stack DLL Client Example is meant to communicate to the CAS BACnet Stack DLL Server Example.");
	CLogging::Fyi("");

	// 1.  Load the callback functions and register the callbacks / hooks
	// 1. Load the callback functions and register the callbacks
	if (!SetupDLLFunctionsAndCallbacks()) {
		CLogging::Error("Failed to load the dll functions and register the callbacks");
		return -2;
	}

	// 2. Test the DLL by using the versioning functions and print the stack version
	CLogging::Fyi("CAS BACnet Stack: v%d.%d.%d.%d", fpGetAPIMajorVersion(), fpGetAPIMinorVersion(), fpGetAPIPatchVersion(), fpGetAPIBuildVersion());
	CLogging::Fyi("");

	// 3. Connect the UDP resource to the BACnet Port
	CLogging::Fyi("Connecting UDP Resource to port=[%u]...", DEFAULT_PORT);
	if (!g_udp.Connect(DEFAULT_PORT)) {
		CLogging::Error("Failed to connect to port=[%u]", DEFAULT_PORT);
		printf("\t Press any key to exit the application...");
		getchar();
		return -3;
	}
	CLogging::Fyi("Connected to port=[%u]", DEFAULT_PORT);

	// 4. Convert the IP Address and Port to a connection string
	CLogging::Fyi("Testing with Server Example at ip=[%s] port=[%u]", ipAddress.c_str(), port);
	if (!CConvert::IPAddressToByteString(ipAddress, connectionString, 6)) {
		CLogging::Error("Failed to convert the ip address to a connection string");
		return -4;
	}
	connectionString[4] = port / 256;
	connectionString[5] = port % 256;
	
	// 5. Setup the device object
	if (!SetupBACnetDevice(BACNET_CLIENT_DEVICE_INSTANCE)) {
		CLogging::Error("Failed to setup the device");
		return -5;
	}

	// 6. Use the Client functions to send messages and the hooks to print out the responses.
	ExampleWhoIs();
	/*
	ExampleWhoIs();
	ExampleWhoHas();
	ExampleReadPropertyDevice();
	ExampleReadPropertyAnalogInput();
	ExampleReadPropertyAnalogOutput();
	ExampleReadPropertyAnalogValue();
	ExampleReadPropertyBinaryInput();
	ExampleReadPropertyBinaryOutput();
	ExampleReadPropertyBinaryValue();
	ExampleReadPropertyMultiStateInput();
	ExampleReadPropertyMultiStateOutput();
	ExampleReadPropertyMultiStateValue();
	ExampleReadPropertyBitstringValue();
	ExampleReadPropertyCharacterstringValue();
	ExampleReadPropertyDateValue();
	ExampleReadPropertyIntegerValue();
	ExampleReadPropertyLargeAnalogValue();
	ExampleReadPropertyOctetstringValue();
	ExampleReadPropertyPositiveIntegerValue();
	ExampleReadPropertyTimeValue();
	ExampleReadPropertyNetworkPort();
	ExampleWritePropertyAnalogValue();
	ExampleWritePropertyAnalogOutput();
	ExampleSubscribeCOV();
	ExampleCreateObjectByType();
	ExampleCreateObjectByTypeWithInitialValues();
	ExampleCreateObjectByIdentifier();
	ExampleCreateObjectByIdentifierWithInitialValues();
	ExampleSendTimeSynchronization();
	ExampleSendUTCTimeSynchronization(); 
	ExampleReadRange();
	ExampleReadRangeByPosition();
	ExampleReadRangeBySequenceNumber();
	ExampleReadRangeByTime();
	ExampleReinitializeDeviceNoPassword();
	ExampleReinitializeDeviceWithPassword();
	ExampleDeviceCommunicationControlNoOptionalParameters();
	ExampleDeviceCommunicationControlWithOptionalParameters();
	*/
	// ExampleReadPropertyAsync();
	// ExampleWritePropertyAsync(); 
	return 0;
}


bool SetupDLLFunctionsAndCallbacks() {
	CLogging::Fyi("Loading DLL Functions...");
	if (!LoadBACnetFunctions()) {
		CLogging::Error("Failed to load the functions");
		return false;
	}

	CLogging::Fyi("Successfully loaded the functions ");
	CLogging::Fyi("");

	CLogging::Fyi("Registering Callback Functions ...");
	if (!LoadDLLCallbackFunctions()) {
		CLogging::Error("Failed to register the callback functions");
		return false;
	}

	CLogging::Fyi("Successfully registered the callback functions");
	CLogging::Fyi("");

	return true;
}

// Sets up a device and one of each object type currently supported by the CAS BACnet Stack
bool SetupBACnetDevice(uint32_t deviceInstance) {
	CLogging::Fyi("Setting up Example Client Device %u...", deviceInstance);

	// Create the Device
	if (!fpAddDevice(deviceInstance)) {
		CLogging::Error("Failed to add Device %u", deviceInstance);
		return false;
	}
	CLogging::Fyi("Created Device %u", deviceInstance);

	// Enable the services that this device supports
	// Some services are mandatory for BACnet devices and are already enabled.
	// These are:
	// Read Property
	// Who Is
	// Who Has
	// 
	// Any other services need to be enabled as below.
	if (!fpSetServiceEnabled(deviceInstance, CASBACnetStackExampleConstants::SERVICE_READ_PROPERTY_MULTIPLE, true)) {
		CLogging::Error("Failed to enabled the ReadPropertyMultiple service=[%u] for Device %u", CASBACnetStackExampleConstants::SERVICE_READ_PROPERTY_MULTIPLE, deviceInstance);
		return false;
	}
	CLogging::Fyi("Enabled ReadPropertyMultiple for Device %u", deviceInstance);

	if (!fpSetServiceEnabled(deviceInstance, CASBACnetStackExampleConstants::SERVICE_I_AM, true)) {
		CLogging::Error("Failed to enabled the IAm service=[%u] for Device %u", CASBACnetStackExampleConstants::SERVICE_I_AM, deviceInstance);
		return false;
	}
	CLogging::Fyi("Enabled IAm for Device %u", deviceInstance);

	if (!fpSetServiceEnabled(deviceInstance, CASBACnetStackExampleConstants::SERVICE_I_HAVE, true)) {
		CLogging::Error("Failed to enabled the IHave service=[%u] for Device %u", CASBACnetStackExampleConstants::SERVICE_I_HAVE, deviceInstance);
		return false;
	}
	CLogging::Fyi("Enabled IHave for Device %u", deviceInstance);

	if (!fpSetServiceEnabled(deviceInstance, CASBACnetStackExampleConstants::SERVICE_CONFIRMED_COV_NOTIFICATION, true)) {
		CLogging::Error("Failed to enabled the ConfirmedCOVNotification service=[%u] for Device %u", CASBACnetStackExampleConstants::SERVICE_CONFIRMED_COV_NOTIFICATION, deviceInstance);
		return false;
	}
	CLogging::Fyi("Enabled ConfirmedCOVNotification for Device %u", deviceInstance);

	if (!fpSetServiceEnabled(deviceInstance, CASBACnetStackExampleConstants::SERVICE_UNCONFIRMED_COV_NOTIFICATION, true)) {
		CLogging::Error("Failed to enabled the UnconfirmedCOVNotification service=[%u] for Device %u", CASBACnetStackExampleConstants::SERVICE_UNCONFIRMED_COV_NOTIFICATION, deviceInstance);
		return false;
	}
	CLogging::Fyi("Enabled UnconfirmedCOVNotification for Device %u", deviceInstance);

	return true;
}

bool LoadDLLCallbackFunctions() {
	// Set up CallBack functions 
	fpRegisterCallbackReceiveMessage(CallbackReceiveMessage);
	fpRegisterCallbackSendMessage(CallbackSendMessage);

	// System Functions
	fpRegisterCallbackGetSystemTime(CallbackGetSystemTime);
		

	fpRegisterHookIAm(HookIAm);
	fpRegisterHookIHave(HookIHave);

	// Hooks for simple PDUs
	fpRegisterHookError(HookError);
	fpRegisterHookReject(HookReject);
	fpRegisterHookAbort(HookAbort);
	fpRegisterHookSimpleAck(HookSimpleAck);

	// Hooks for different ComplexAck PDUs
	fpRegisterHookPropertyBitString(HookPropertyBitString);
	fpRegisterHookPropertyBool(HookPropertyBool);
	fpRegisterHookPropertyCharString(HookPropertyCharString);
	fpRegisterHookPropertyDate(HookPropertyDate);
	fpRegisterHookPropertyDouble(HookPropertyDouble);
	fpRegisterHookPropertyEnum(HookPropertyEnum);
	fpRegisterHookPropertyNull(HookPropertyNull);
	fpRegisterHookPropertyObjectIdentifier(HookPropertyObjectIdentifier);
	fpRegisterHookPropertyOctString(HookPropertyOctString);
	fpRegisterHookPropertyInt(HookPropertyInt);
	fpRegisterHookPropertyReal(HookPropertyReal);
	fpRegisterHookPropertyTime(HookPropertyTime);
	fpRegisterHookPropertyUInt(HookPropertyUInt);
	return true;
}

uint16_t CallbackReceiveMessage(uint8_t * message, const uint16_t maxMessageLength, uint8_t * receivedConnectionString, const uint8_t maxConnectionStringLength, uint8_t * receivedConnectionStringLength, uint8_t * networkType) {
	if (message == NULL || maxMessageLength == 0) {
		CLogging::Error("Invalid input buffer");
		return 0;
	}
	if (receivedConnectionString == NULL || maxConnectionStringLength == 0) {
		CLogging::Error("Invalid connection string buffer");
		return 0;
	}
	if (maxConnectionStringLength < 6) {
		CLogging::Error("Not enough space for a UDP connection string");
		return 0;
	}

	char ipAddress[32];
	uint16_t port = 0;

	int bytesRead = g_udp.GetMessage(message, maxMessageLength, ipAddress, &port);
	if (bytesRead > 0) {
		CASBACnetStack::CEndianness::ToBigEndian(&port, sizeof(uint16_t));
		CLogging::Fyi("Received message from [%s:%u]", ipAddress, port);

		// Convert the IP Address to the connection string
		if (!CConvert::IPAddressToByteString(ipAddress, receivedConnectionString, maxConnectionStringLength)) {
			CLogging::Error("Failed to convert the ip address into a connectionString");
			return 0;
		}
		receivedConnectionString[4] = port / 256;
		receivedConnectionString[5] = port % 256;

		*receivedConnectionStringLength = 6;
		*networkType = CASBACnetStack::BACnetPacket::NetworkType_IP;

		if (fpDecodeAsXML((char*)message, bytesRead, xmlRenderBuffer, MAX_XML_RENDER_BUFFER_LENGTH) > 0) {
			CLogging::LogBuffer(CLogging::DEBUG_LEVEL_IMPORTANT, xmlRenderBuffer);
			memset(xmlRenderBuffer, 0, MAX_XML_RENDER_BUFFER_LENGTH);
		}
	}

	return bytesRead;
}

uint16_t CallbackSendMessage(const uint8_t * message, const uint16_t messageLength, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, bool broadcast) {
	if (message == NULL || messageLength == 0) {
		printf("Nothing to send\n");
		return 0;
	}
	if (connectionString == NULL || connectionStringLength == 0) {
		printf("No connection string\n");
		return 0;
	}

	// Verify Network Type
	if (networkType != CASBACnetStack::BACnetPacket::NetworkType_IP) {
		printf("Message for different network\n");
		return 0;
	}

	// Prepare the IP Address
	char ipAddress[32];
	if (broadcast) {
		// This example will simply assume a Network Mask of 255.255.255.0
		// When implementing, query the network card to get the proper mask and apply to get the
		// Broadcast address
		snprintf(ipAddress, 32, "%u.%u.%u.255", connectionString[0], connectionString[1], connectionString[2]);
	}
	else {
		snprintf(ipAddress, 32, "%u.%u.%u.%u", connectionString[0], connectionString[1], connectionString[2], connectionString[3]);
	}
	
	uint16_t port = 0;
	port += connectionString[4] * 256;
	port += connectionString[5];

	CLogging::Fyi("Sending message to [%s:%u]", ipAddress, port);

	if (!g_udp.SendMessage(ipAddress, port, (unsigned char *)message, messageLength)) {
		printf("Failed to send message\n");
		return 0;
	}

	if (fpDecodeAsXML((char*)message, messageLength, xmlRenderBuffer, MAX_XML_RENDER_BUFFER_LENGTH) > 0) {
		CLogging::LogBuffer(CLogging::DEBUG_LEVEL_IMPORTANT, xmlRenderBuffer);
		memset(xmlRenderBuffer, 0, MAX_XML_RENDER_BUFFER_LENGTH);
	}

	return messageLength;
}

// System Functions
time_t CallbackGetSystemTime() {
	return time(0);
}

// Hooks for unconfirmed request
void HookIAm(const uint32_t deviceIdentifier, const uint32_t maxApduLengthAccepted, const uint8_t segmentationSupported, const uint16_t vendorIdentifier, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength) {
	// Extract some of the values before printing
	uint16_t fromPort = 0;
	fromPort += (connectionString[4] * 256);
	fromPort += (connectionString[5]);

	CLogging::Fyi("Received IAm from IP=[%u.%u.%u.%u:%u] by Device with ID=[%u] with MaxAPDULength=[%u], SegmentationSupported=[%u], VendorId=[%u]", connectionString[0], connectionString[1], connectionString[2], connectionString[3], fromPort, deviceIdentifier, maxApduLengthAccepted, segmentationSupported, vendorIdentifier);
	if (network == 0) {
		CLogging::Fyi("Device ID=[%u] is on Network=[%u]", deviceIdentifier, network);
	}
	else {
		std::string sadr;
		char sadrByte[3]; 
		if (sourceAddress != NULL && sourceAddressLength != 0) {
			for (size_t i = 0; i < sourceAddressLength; i++) {
				snprintf(sadrByte, 3, "0x%02x", sourceAddress[i]);
			}
		}

		CLogging::Fyi("Device ID=[%u] is on Network=[%u] %s%s", deviceIdentifier, network, sourceAddressLength == 0 ? "" : "with SADR of 0x ", sourceAddressLength == 0 ? "" : sadr.c_str());
	}
}

void HookIHave(const uint32_t deviceIdentifier, const uint16_t objectType, const uint32_t objectInstance, const char * objectName, const uint32_t objectNameLength, const uint8_t objectNameEncoding, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength) {
	// Extract some of the values before printing
	uint16_t fromPort = 0;
	fromPort += (connectionString[4] * 256);
	fromPort += (connectionString[5]);

	std::string oName = std::string(objectName, objectNameLength);

	CLogging::Fyi("Received IHave from IP=[%u.%u.%u.%u:%u] by Device with ID=[%u] with an object of type=[%u], instance=[%u] and name=[%s]", connectionString[0], connectionString[1], connectionString[2], connectionString[3], fromPort, deviceIdentifier, objectType, objectInstance, oName.c_str());
	if (network == 0) {
		CLogging::Fyi("Device ID=[%u] is on Network=[%u]", deviceIdentifier, network);
	}
	else {
		std::string sadr;
		char sadrByte[3];
		if (sourceAddress != NULL && sourceAddressLength != 0) {
			for (size_t i = 0; i < sourceAddressLength; i++) {
				snprintf(sadrByte, 3, "0x%02x", sourceAddress[i]);
			}
		}

		CLogging::Fyi("Device ID=[%u] is on Network=[%u] %s%s", deviceIdentifier, network, sourceAddressLength == 0 ? "" : "with SADR of 0x ", sourceAddressLength == 0 ? "" : sadr.c_str());
	}
}

// Hooks for simple PDUs
void HookError(const uint8_t originalInvokeId, const uint32_t errorChoice, const uint32_t errorClass, const uint32_t errorCode, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength, const bool useObjectProperty, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier) {
	(void)connectionString;
	(void)connectionStringLength;
	(void)networkType;
	(void)network;
	(void)sourceAddress;
	(void)sourceAddressLength;
	(void)useObjectProperty;
	(void)objectType;
	(void)objectInstance;
	(void)propertyIdentifier;

	CLogging::Fyi("Received Error with originalInvokeId=[%u] for errorChoice=[%u], errorClass=[%u], errorCode=[%u]", originalInvokeId, errorChoice, errorClass, errorCode);
}

void HookReject(const uint8_t originalInvokeId, const uint32_t rejectReason, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength) {
	(void)connectionString;
	(void)connectionStringLength;
	(void)networkType;
	(void)network;
	(void)sourceAddress;
	(void)sourceAddressLength;

	CLogging::Fyi("Received Reject with originalInvokeId=[%u] and rejectReason=[%u]", originalInvokeId, rejectReason);
}

void HookAbort(const uint8_t originalInvokeId, const bool sentByServer, const uint32_t abortReason, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength) {
	(void)connectionString;
	(void)connectionStringLength;
	(void)networkType;
	(void)network;
	(void)sourceAddress;
	(void)sourceAddressLength;

	if (sentByServer) {
		CLogging::Fyi("Received Abort from a server device with originalInvokeId=[%u] and abortReason=[%u]", originalInvokeId, abortReason);
	}
	else {
		CLogging::Fyi("Received Abort from a client device with originalInvokeId=[%u] and abortReason=[%u]", originalInvokeId, abortReason);
	}
}

void HookSimpleAck(const uint8_t originalInvokeId, const uint32_t serverAckChoice, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength) {
	(void)connectionString;
	(void)connectionStringLength;
	(void)networkType;
	(void)network;
	(void)sourceAddress;
	(void)sourceAddressLength;

	CLogging::Fyi("Received SimpleAck with originalInvokeId=[%u] and serverAckChoice=[%u]", originalInvokeId, serverAckChoice);
}

void HookTimeout(const uint8_t originalInvokeId, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength) {
	(void)connectionString;
	(void)connectionStringLength;
	(void)networkType;
	(void)network;
	(void)sourceAddress;
	(void)sourceAddressLength;

	CLogging::Fyi("ConfirmedRequest with originalInvokeId=[%u] timed out", originalInvokeId);
}

// Hooks for different ComplexAck PDUs
void HookPropertyBitString(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const bool * value, const uint32_t length, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength) {
	(void)connectionString;
	(void)connectionStringLength;
	(void)networkType;
	(void)network;
	(void)sourceAddress;
	(void)sourceAddressLength;

	std::string bitstringValue = "";
	bitstringValue.append("{");
	for (size_t i = 0; i < length; i++) {
		if (i != 0) {
			bitstringValue.append(",");
		}
		bitstringValue.append(value[i] ? "1" : "0");
	}
	bitstringValue.append("}");

	if (!usePropertyArrayIndex) {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u] and value=[%s]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, bitstringValue.c_str());
	}
	else {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u], propertyArrayIndex=[%u] and value=[%s]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, propertyArrayIndex, bitstringValue.c_str());
	}
}

void HookPropertyBool(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const bool value, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength) {
	(void)connectionString;
	(void)connectionStringLength;
	(void)networkType;
	(void)network;
	(void)sourceAddress;
	(void)sourceAddressLength;

	if (!usePropertyArrayIndex) {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u] and value=[%s]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, value ? "true" : "false");
	}
	else {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u], propertyArrayIndex=[%u] and value=[%s]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, propertyArrayIndex, value ? "true" : "false");
	}
}

void HookPropertyCharString(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const char * value, const uint32_t length, const uint8_t encoding, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength) {
	(void)connectionString;
	(void)connectionStringLength;
	(void)networkType;
	(void)network;
	(void)sourceAddress;
	(void)sourceAddressLength;


	if (!usePropertyArrayIndex) {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u] and value=[%s]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, length == 0 ? "" : value);
	}
	else {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u], propertyArrayIndex=[%u] and value=[%s]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, propertyArrayIndex, length == 0 ? "" : value);
	}
}

void HookPropertyDate(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const uint8_t year, const uint8_t month, const uint8_t day, const uint8_t weekday, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength) {
	(void)connectionString;
	(void)connectionStringLength;
	(void)networkType;
	(void)network;
	(void)sourceAddress;
	(void)sourceAddressLength;

	if (!usePropertyArrayIndex) {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u] and value=[%u/%u/%u (%u)]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, year + 1900, month, day, weekday);
	}
	else {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u], propertyArrayIndex=[%u] and value=[%u/%u/%u (%u)]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, propertyArrayIndex, year + 1900, month, day, weekday);
	}
}

void HookPropertyDouble(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const double value, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength) {
	(void)connectionString;
	(void)connectionStringLength;
	(void)networkType;
	(void)network;
	(void)sourceAddress;
	(void)sourceAddressLength;

	if (!usePropertyArrayIndex) {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u] and value=[%f]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, value);
	}
	else {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u], propertyArrayIndex=[%u] and value=[%f]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, propertyArrayIndex, value);
	}
}

void HookPropertyEnum(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const uint32_t value, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength) {
	(void)connectionString;
	(void)connectionStringLength;
	(void)networkType;
	(void)network;
	(void)sourceAddress;
	(void)sourceAddressLength;

	if (!usePropertyArrayIndex) {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u] and value=[%u]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, value);
	}
	else {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u], propertyArrayIndex=[%u] and value=[%u]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, propertyArrayIndex, value);
	}
}

void HookPropertyNull(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength) {
	(void)connectionString;
	(void)connectionStringLength;
	(void)networkType;
	(void)network;
	(void)sourceAddress;
	(void)sourceAddressLength;

	if (!usePropertyArrayIndex) {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u] and value=[null]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier);
	}
	else {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u], propertyArrayIndex=[%u] and value=[null]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, propertyArrayIndex);
	}
}

void HookPropertyObjectIdentifier(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const uint16_t objectTypeValue, const uint32_t objectInstanceValue, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength) {
	(void)connectionString;
	(void)connectionStringLength;
	(void)networkType;
	(void)network;
	(void)sourceAddress;
	(void)sourceAddressLength;

	if (!usePropertyArrayIndex) {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u] and value=[%u (%u)]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, objectTypeValue, objectInstanceValue);
	}
	else {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u], propertyArrayIndex=[%u] and value=[%u (%u)]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, propertyArrayIndex, objectTypeValue, objectInstanceValue);
	}
}

void HookPropertyOctString(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const uint8_t * value, const uint32_t length, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength) {
	(void)connectionString;
	(void)connectionStringLength;
	(void)networkType;
	(void)network;
	(void)sourceAddress;
	(void)sourceAddressLength;

	std::string octStringValue = "{";
	char octStringByte[3];
	for (size_t i = 0; i < length; i++) {
		memset(octStringByte, 0, 3);
		snprintf(octStringByte, 3, "%02X", value[i]);
		octStringValue.append(octStringByte);
	}
	octStringValue.append("}");
	
	if (!usePropertyArrayIndex) {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u] and value=[%s]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, octStringValue.c_str());
	}
	else {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u], propertyArrayIndex=[%u] and value=[%s]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, propertyArrayIndex, octStringValue.c_str());
	}
}

void HookPropertyInt(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const int32_t value, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength) {
	(void)connectionString;
	(void)connectionStringLength;
	(void)networkType;
	(void)network;
	(void)sourceAddress;
	(void)sourceAddressLength;

	if (!usePropertyArrayIndex) {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u] and value=[%d]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, value);
	}
	else {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u], propertyArrayIndex=[%u] and value=[%d]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, propertyArrayIndex, value);
	}
}

void HookPropertyReal(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const float value, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength) {
	(void)connectionString;
	(void)connectionStringLength;
	(void)networkType;
	(void)network;
	(void)sourceAddress;
	(void)sourceAddressLength;

	if (!usePropertyArrayIndex) {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u] and value=[%f]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, value);
	}
	else {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u], propertyArrayIndex=[%u] and value=[%f]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, propertyArrayIndex, value);
	}

	// Check for Present Value of the Analog Value
	if (objectType == CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_VALUE && objectInstance == 2) {
		if (propertyIdentifier == CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE) {
			analogValuePresentValue = value;
		}
	}
}

void HookPropertyTime(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const uint8_t hour, const uint8_t minute, const uint8_t second, const uint8_t hundrethSecond, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength) {
	(void)connectionString;
	(void)connectionStringLength;
	(void)networkType;
	(void)network;
	(void)sourceAddress;
	(void)sourceAddressLength;

	if (!usePropertyArrayIndex) {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u] and value=[%u:%u:%u:%u]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, hour, minute, second, hundrethSecond);
	}
	else {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u], propertyArrayIndex=[%u] and value=[%u:%u:%u:%u]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, propertyArrayIndex, hour, minute, second, hundrethSecond);
	}
}

void HookPropertyUInt(const uint32_t originalInvokeId, const uint8_t service, const uint16_t objectType, const uint32_t objectInstance, const uint32_t propertyIdentifier, const bool usePropertyArrayIndex, const uint32_t propertyArrayIndex, const uint32_t value, const uint8_t * connectionString, const uint8_t connectionStringLength, const uint8_t networkType, const uint16_t network, const uint8_t * sourceAddress, const uint8_t sourceAddressLength) {
	(void)connectionString;
	(void)connectionStringLength;
	(void)networkType;
	(void)network;
	(void)sourceAddress;
	(void)sourceAddressLength;

	if (!usePropertyArrayIndex) {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u] and value=[%u]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, value);
	}
	else {
		CLogging::Fyi("Received service=[%u] with originalInvokedId=[%u] for objectType=[%u], objectInstance=[%u], propertyIdentifier=[%u], propertyArrayIndex=[%u] and value=[%u]", service, originalInvokeId, objectType, objectInstance, propertyIdentifier, propertyArrayIndex, value);
	}
}

bool ParseCommandLine(int argc, char ** argv) {
	// Skip the first argument as that is the application name
	for (int i = 1; i < argc; i++) {
		// Iterate through the arguments to extract the IP Address and Port
		std::string arg = std::string(argv[i]);
		// Search for ip address argument
		if (arg.find("-i") != std::string::npos) {
			ipAddress = arg.substr(2);
		}
		else if (arg.find("-p") != std::string::npos) {
			port = ToValue<uint16_t>(arg.substr(2));
		}
		else {
			CLogging::Error("Invalid command line argument. Usage: [app name] -i{IPAddress} -p{Port}.");
			return false;
		}
	}

	return true;
}

void WaitForResponse(unsigned int timeout /*=3*/) {
	CLogging::Fyi("Entering Loop for response...");
	time_t expireTime = time(0) + timeout;
	while (time(0) < expireTime) {
		fpLoop();
	}
	printf("\n");
}

void ExampleReadPropertyAnalogInputPresentValue() {
	CLogging::Fyi("Sending ReadProperty for the Analog Input Present Value...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_INPUT, 0, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleWhoIs() {
	// 1. Send the message
	CLogging::Fyi("Sending WhoIs with no range...");
	fpSendWhoIs(connectionString, 6, 0, true, 0, NULL, 0);

	// 2. Enter a 3 second loop giving time for the hooks to be called if a message responds.
	WaitForResponse();

	CLogging::Fyi("Sending WhoIs with range...");
	fpSendWhoIsWithLimits(389900, 389999, connectionString, 6, 0, true, 0, NULL, 0);

	WaitForResponse();

	CLogging::Fyi("Sending WhoIs to specific network");
	fpSendWhoIs(connectionString, 6, 0, true, 38900, NULL, 0);

	WaitForResponse();

	CLogging::Fyi("Sending WhoIs to broadcast network");
	fpSendWhoIs(connectionString, 6, 0, true, 65535, NULL, 0);

	WaitForResponse();

}

void ExampleWhoHas() {
	CLogging::Fyi("Sending WhoHas for Object Identifier...");
	fpSendWhoHasWithObjectIdentifier(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_INPUT, 0, connectionString, 6, 0, true, 0, NULL, 0);
	WaitForResponse();

	CLogging::Fyi("Sending WhoHas for Object Name...");
	std::string objectName = "AnalogInput_Red";
	fpSendWhoHasWithObjectName(objectName.c_str(), objectName.size(), connectionString, 6, 0, true, 0, NULL, 0);
	WaitForResponse();
}

void ExampleReadPropertyDevice() {
	CLogging::Fyi("Sending ReadPropertyMultiple All for the Device...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_DEVICE, 389999, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ALL, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleReadPropertyAnalogInput() {
	CLogging::Fyi("Sending ReadPropertyMultiple All for the Analog Input...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_INPUT, 0, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ALL, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleReadPropertyAnalogOutput() {
	CLogging::Fyi("Sending ReadPropertyMultiple All for the Analog Output...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ALL, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleReadPropertyAnalogValue() {
	CLogging::Fyi("Sending ReadPropertyMultiple All for the Analog Value...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_VALUE, 2, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ALL, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleReadPropertyBinaryInput() {
	CLogging::Fyi("Sending ReadPropertyMultiple All for the Binary Input...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_BINARY_INPUT, 3, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ALL, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleReadPropertyBinaryOutput() {
	CLogging::Fyi("Sending ReadPropertyMultiple All for the Binary Output...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_BINARY_OUTPUT, 4, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ALL, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleReadPropertyBinaryValue() {
	CLogging::Fyi("Sending ReadPropertyMultiple All for the Binary Value...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_BINARY_VALUE, 5, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ALL, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleReadPropertyMultiStateInput() {
	CLogging::Fyi("Sending ReadPropertyMultiple All for the Multi State Input...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_MULTI_STATE_INPUT, 13, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ALL, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleReadPropertyMultiStateOutput() {
	CLogging::Fyi("Sending ReadPropertyMultiple All for the Multi State Output...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_MULTI_STATE_OUTPUT, 14, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ALL, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleReadPropertyMultiStateValue() {
	CLogging::Fyi("Sending ReadPropertyMultiple All for the Multi State Value...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_MULTI_STATE_VALUE, 19, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ALL, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleReadPropertyBitstringValue() {
	CLogging::Fyi("Sending ReadPropertyMultiple All for the Bitstring Value...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_BITSTRING_VALUE, 39, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ALL, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleReadPropertyCharacterstringValue() {
	CLogging::Fyi("Sending ReadPropertyMultiple All for the Characterstring Value...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_CHARACTERSTRING_VALUE, 40, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ALL, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleReadPropertyDateValue() {
	CLogging::Fyi("Sending ReadPropertyMultiple All for the Date Value...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_DATE_VALUE, 42, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ALL, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleReadPropertyIntegerValue() {
	CLogging::Fyi("Sending ReadPropertyMultiple All for the Integer Value...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_INTEGER_VALUE, 45, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ALL, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleReadPropertyLargeAnalogValue() {
	CLogging::Fyi("Sending ReadPropertyMultiple All for the Large Analog Value...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_LARGE_ANALOG_VALUE, 46, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ALL, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleReadPropertyOctetstringValue() {
	CLogging::Fyi("Sending ReadPropertyMultiple All for the Octetstring Value...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_OCTETSTRING_VALUE, 47, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ALL, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleReadPropertyPositiveIntegerValue() {
	CLogging::Fyi("Sending ReadPropertyMultiple All for the Positive Integer Value...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_POSITIVE_INTEGER_VALUE, 48, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ALL, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleReadPropertyTimeValue() {
	CLogging::Fyi("Sending ReadPropertyMultiple All for the Time Value...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_TIME_VALUE, 50, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ALL, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleWritePropertyAnalogValue() {
	CLogging::Fyi("Sending ReadProperty for the Present Value of Analog Value 2...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_VALUE, 2, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending WriteProperty to the Present Value of Analog Value 2...");
	
	fpBuildWriteProperty(4, "1.0", 3, CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_VALUE, 2, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, false, 0, false, 16);
	fpSendWriteProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending ReadProperty for the Present Value of Analog Value 2 to verify the write...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_VALUE, 2, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();
}


void ExampleWritePropertyAnalogOutput() {
	CLogging::Fyi("Sending ReadProperty for the Relinquish Default of Analog Output 1...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_RELINQUISH_DEFAULT, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending ReadProperty for the Present Value of Analog Output 1...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending ReadProperty for the Priority Array of Analog Output 1...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRIORITY_ARRAY, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending WriteProperty to the Present Value of Analog Output 1 at priority 14...");
	fpBuildWriteProperty(4, "14.0", 4, CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, false, 0, true, 14);
	fpSendWriteProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending ReadProperty to confirm the Present Value of Analog Output 1...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending ReadProperty to confirm the Priority Array of Analog Output 1...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRIORITY_ARRAY, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending WriteProperty to the Present Value of Analog Output 1 at priority 8...");
	fpBuildWriteProperty(4, "8.0", 3, CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, false, 0, true, 8);
	fpSendWriteProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending ReadProperty to confirm the Present Value of Analog Output 1...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending ReadProperty to confirm the Priority Array of Analog Output 1...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRIORITY_ARRAY, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending WriteProperty to the Present Value of Analog Output 1 at priority 2...");
	fpBuildWriteProperty(4, "2.0", 3, CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, false, 0, true, 2);
	fpSendWriteProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending ReadProperty to confirm the Present Value of Analog Output 1...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending ReadProperty to confirm the Priority Array of Analog Output 1...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRIORITY_ARRAY, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending WriteProperty to the Present Value of Analog Output 1 to relinquish priority 2...");
	fpBuildWriteProperty(0, "", 0, CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, false, 0, true, 2);
	fpSendWriteProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending ReadProperty to confirm the Present Value of Analog Output 1...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending ReadProperty to confirm the Priority Array of Analog Output 1...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRIORITY_ARRAY, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending WriteProperty to the Present Value of Analog Output 1 to relinquish priority 8...");
	fpBuildWriteProperty(0, "", 0, CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, false, 0, true, 8);
	fpSendWriteProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending ReadProperty to confirm the Present Value of Analog Output 1...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending ReadProperty to confirm the Priority Array of Analog Output 1...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRIORITY_ARRAY, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending WriteProperty to the Present Value of Analog Output 1 to relinquish priority 14...");
	fpBuildWriteProperty(0, "", 0, CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, false, 0, true, 14);
	fpSendWriteProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending ReadProperty to confirm the Present Value of Analog Output 1...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending ReadProperty to confirm the Priority Array of Analog Output 1...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, 1, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRIORITY_ARRAY, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();
}

void ExampleSubscribeCOV() {
	CLogging::Fyi("Sending ReadProperty for the Present Value of Analog Value 2...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_VALUE, 2, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending SubscribeCOV for Analog Value 2...");
	fpSendSubscribeCOV(&invokeId, 1, CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_VALUE, 2, false, 30, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();

	CLogging::Fyi("Sending ReadProperty for the Active COV Subscriptions of Device...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_DEVICE, 389999, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ACTIVE_COV_SUBSCRIPTIONS, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	
	WaitForResponse();
}

void ExampleReadPropertyNetworkPort() {
	CLogging::Fyi("Sending ReadPropertyMultiple All for the Network Port...");
	fpBuildReadProperty(CASBACnetStackExampleConstants::OBJECT_TYPE_NETWORK_PORT, 56, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_ALL, false, 0);
	fpSendReadProperty(&invokeId, connectionString, 6, 0, 0, NULL, 0);

	WaitForResponse();
}

void ExampleCreateObjectByType() {
	CLogging::Fyi("Sending CreateObject using object type for an analog input");
	if (!fpSendCreateObjectByType(&invokeId, CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_INPUT, connectionString, 6, 0, 0, NULL, 0)) {
		CLogging::Error("Failed to send create object message");
		return;
	}

	WaitForResponse();
}

void ExampleCreateObjectByTypeWithInitialValues() {
	CLogging::Fyi("Sending CreateObject using object type for an analog input with initial value for cov increment");
	std::string covIncrement = "0.5f";
	if (!fpBuildCreateObjectInitialPropertyValues(CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_COV_INCREMENT, CASBACnetStackExampleConstants::BACNET_DATATYPE_REAL, covIncrement.c_str(), covIncrement.size())) {
		CLogging::Error("Failed to add initial property value for cov increment");
		return;
	}

	if (!fpSendCreateObjectByType(&invokeId, CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_INPUT, connectionString, 6, 0, 0, NULL, 0)) {
		CLogging::Error("Failed to send create object message");
		return;
	}

	WaitForResponse();
}

void ExampleCreateObjectByIdentifier() {
	CLogging::Fyi("Sending CreateObject using object identifier for an analog input 123456");
	if (!fpSendCreateObjectByIdentifier(&invokeId, CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_INPUT, 123456, connectionString, 6, 0, 0, NULL, 0)) {
		CLogging::Error("Failed to send create object message");
		return;
	}
	
	WaitForResponse();
}
void ExampleCreateObjectByIdentifierWithInitialValues() {
	CLogging::Fyi("Sending CreateObject using object identifier for an analog input 123457 with initial value for cov increment and object name");
	std::string initialCovIncrement = "0.5f";
	std::string initialObjectName = "Hello World";
	if (!fpBuildCreateObjectInitialPropertyValues(CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_COV_INCREMENT, CASBACnetStackExampleConstants::BACNET_DATATYPE_REAL, initialCovIncrement.c_str(), initialCovIncrement.size())) {
		CLogging::Error("Failed to add initial property value for cov increment");
		return;
	}

	if (!fpBuildCreateObjectInitialPropertyValues(CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_OBJECT_NAME, CASBACnetStackExampleConstants::BACNET_DATATYPE_CHARACTER_STRING, initialObjectName.c_str(), initialObjectName.size())) {
		CLogging::Error("Failed to add initial property value for object type");
		return;
	}
	
	if (!fpSendCreateObjectByIdentifier(&invokeId, CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_INPUT, 123457, connectionString, 6, 0, 0, NULL, 0)) {
		CLogging::Error("Failed to send create object message");
		return;
	}

	WaitForResponse();
}

void ExampleDeleteObject() {
	CLogging::Fyi("Sending DeleteObject");

	if (!fpSendDeleteObject(&invokeId, CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_INPUT, 123456, connectionString, 6, 0, 0, NULL, 0)) {
		CLogging::Error("Failed to send delete object message");
		return;
	}

	WaitForResponse();
}

void ExampleSendTimeSynchronization() {
	CLogging::Fyi("Sending SendTimeSynchronization");


	const uint8_t year = (2018-1900); // 1900 = 0, 2018 = 118, 2155 = 255 
	const uint8_t month = 10;
	const uint8_t day = 3;
	const uint8_t weekday = 3;
	const uint8_t hour = 12;
	const uint8_t minute = 0;
	const uint8_t second = 0;
	const uint8_t hundrethSeconds = 0;

	if (!fpSendTimeSynchronization(year, month, day, weekday, hour, minute, second, hundrethSeconds, connectionString, 6, 0, false, 0, NULL, 0)) {
		CLogging::Error("Failed to Send Time Synchronization message");
		return;
	}

	WaitForResponse();
}

void ExampleSendUTCTimeSynchronization() {
	CLogging::Fyi("Sending SendUTCTimeSynchronization");


	const uint8_t year = (2018 - 1900); // 1900 = 0, 2018 = 118, 2155 = 255 
	const uint8_t month = 10;
	const uint8_t day = 3;
	const uint8_t weekday = 3;
	const uint8_t hour = 12;
	const uint8_t minute = 0;
	const uint8_t second = 0;
	const uint8_t hundrethSeconds = 0;

	if (!fpSendUTCTimeSynchronization(year, month, day, weekday, hour, minute, second, hundrethSeconds, connectionString, 6, 0, false, 0, NULL, 0)) {
		CLogging::Error("Failed to Send UTD Time Synchronization message");
		return;
	}

	WaitForResponse();
}

void ExampleReadRange() {
	CLogging::Fyi("Sending ReadRange");
	if (!fpSendReadRange(&invokeId, CASBACnetStackExampleConstants::OBJECT_TYPE_TREND_LOG, 20, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_LOG_BUFFER, false, 0, connectionString, 6, 0, 0, NULL, 0)) {
		CLogging::Error("Failed to send read range message");
		return;
	}

	WaitForResponse();
}

void ExampleReadRangeByPosition() {
	CLogging::Fyi("Sending ReadRange");
	if (!fpSendReadRangeByPosition(&invokeId, CASBACnetStackExampleConstants::OBJECT_TYPE_TREND_LOG, 20, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_LOG_BUFFER, false, 0, 0, 10, connectionString, 6, 0, 0, NULL, 0)) {
		CLogging::Error("Failed to send read range by position message");
		return;
	}

	WaitForResponse();
}

void ExampleReadRangeBySequenceNumber() {
	CLogging::Fyi("Sending ReadRange");
	if (!fpSendReadRangeBySequenceNumber(&invokeId, CASBACnetStackExampleConstants::OBJECT_TYPE_TREND_LOG, 20, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_LOG_BUFFER, false, 0, 10, 10, connectionString, 6, 0, 0, NULL, 0)) {
		CLogging::Error("Failed to send read range by sequence number message");
		return;
	}

	WaitForResponse();
}

void ExampleReadRangeByTime() {
	CLogging::Fyi("Sending ReadRange");
	if (!fpSendReadRangeByTime(&invokeId, CASBACnetStackExampleConstants::OBJECT_TYPE_TREND_LOG, 20, CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_LOG_BUFFER, false, 0, 118, 9, 27, 4, 8, 0, 0, 0, 10, connectionString, 6, 0, 0, NULL, 0)) {
		CLogging::Error("Failed to send read range by time message");
		return;
	}

	WaitForResponse();
}

void ExampleReinitializeDeviceNoPassword() {
	CLogging::Fyi("Sending ReinitializeDevice with no password");
	if (!fpSendReinitializeDevice(&invokeId, CASBACnetStackExampleConstants::BACNET_REINITIALIZED_STATE_OF_DEVICE_ACTIVATECHANGES, NULL, 0 ,connectionString, 6, 0, 0, NULL, 0)) {
		CLogging::Error("Failed to send reinitialize device with no password");
		return;
	}

	WaitForResponse();
}

void ExampleReinitializeDeviceWithPassword() {
	CLogging::Fyi("Sending ReinitializeDevice with password");
	if (!fpSendReinitializeDevice(&invokeId, CASBACnetStackExampleConstants::BACNET_REINITIALIZED_STATE_OF_DEVICE_WARMSTART, "12345", 5, connectionString, 6, 0, 0, NULL, 0)) {
		CLogging::Error("Failed to send reinitialize device with password");
		return;
	}

	WaitForResponse();
}

void ExampleDeviceCommunicationControlNoOptionalParameters() {
	CLogging::Fyi("Sending DeviceCommunicationControl with no optional parameters");
	if (!fpSendDeviceCommunicationControl(&invokeId, CASBACnetStackExampleConstants::BACNET_ENABLE_DISABLE_DISABLE_INITIATION, false, 0, NULL, 0, connectionString, 6, 0, 0, NULL, 0)) {
		CLogging::Error("Failed to send DeviceCommunicationControl with no optional parameters");
		return;
	}

	WaitForResponse();
}

void ExampleDeviceCommunicationControlWithOptionalParameters() {
	CLogging::Fyi("Sending DeviceCommunicationControl with optional parameters");
	if (!fpSendDeviceCommunicationControl(&invokeId, CASBACnetStackExampleConstants::BACNET_ENABLE_DISABLE_DISABLE, true, 600, "12345", 5, connectionString, 6, 0, 0, NULL, 0)) {
		CLogging::Error("Failed to send DeviceCommunicationControl with optional parameters");
		return;
	}

	WaitForResponse();
}

void ExampleReadPropertyAsync() {
	CLogging::Fyi("Sending SendReadPropertyAsync");
	
	/*
	*  120 bits (15 BYTES)
	*  *-[0]-----------*-[1][2]--------------*-[3][4][5][6]------------*-[7][8][9][10]---------------*-[11][12][13][14]------------*
	*  | uint8_t flags | uint16_t objectType | uint32_t objectInstance | uint32_t propertyIdentifier | uint32_t propertyArrayIndex |
	*  *---------------*---------------------*-------------------------*-----------------------------*-----------------------------*
	*/

	uint8_t readPropertyBuffer[15 * 3];
	memset(readPropertyBuffer, 0, 15 * 3);

	// 0 
	uint8_t offset = 0; 
	uint32_t objectInstance = 0; 
	readPropertyBuffer[0 + (offset * 15)] = 0x00; // Flags 
	memcpy(readPropertyBuffer + 1 + (offset * 15), &CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_INPUT, sizeof(uint16_t)); // objectType
	memcpy(readPropertyBuffer + 3 + (offset * 15), &objectInstance, sizeof(uint32_t)); // objectInstance
	memcpy(readPropertyBuffer + 7 + (offset * 15), &CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, sizeof(uint32_t)); // propertyIdentifier

	// 1 
	offset = 1;
	objectInstance = 3;
	readPropertyBuffer[0 + (offset * 15)] = 0x00; // Flags 
	memcpy(readPropertyBuffer + 1 + (offset * 15), &CASBACnetStackExampleConstants::OBJECT_TYPE_BINARY_INPUT, sizeof(uint16_t)); // objectType
	memcpy(readPropertyBuffer + 3 + (offset * 15), &objectInstance, sizeof(uint32_t)); // objectInstance
	memcpy(readPropertyBuffer + 7 + (offset * 15), &CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, sizeof(uint32_t)); // propertyIdentifier

	// 2
	offset = 2;
	objectInstance = 13;
	uint32_t propertyArrayIndex = 3; 
	readPropertyBuffer[0 + (offset * 15)] = 0x01; // Flags. 0x01 == use propertyArrayIndex
	memcpy(readPropertyBuffer + 1 + (offset * 15), &CASBACnetStackExampleConstants::OBJECT_TYPE_MULTI_STATE_INPUT, sizeof(uint16_t)); // objectType
	memcpy(readPropertyBuffer + 3 + (offset * 15), &objectInstance, sizeof(uint32_t)); // objectInstance
	memcpy(readPropertyBuffer + 7 + (offset * 15), &CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, sizeof(uint32_t)); // propertyIdentifier
	memcpy(readPropertyBuffer + 11 + (offset * 15), &propertyArrayIndex, sizeof(uint32_t)); // propertyArrayIndex

	if (!fpSendReadPropertyAsync(&invokeId, connectionString, 6, 0, 0, NULL, 0, 3, (void *)readPropertyBuffer)) {
		CLogging::Error("Failed to send SendReadPropertyAsync");
		return;
	}

	WaitForResponse();
}


void ExampleWritePropertyAsync() {
	CLogging::Fyi("Sending SendWritePropertyAsync");

	char * WriteValueOne = "1.234";
	char * WriteValueTwo = "1";
	char * WriteValueThree = "7";


	/*
	 *
     *  *-[0]-----------*-[1][2]--------------*-[3][4][5][6]------------*-[7][8][9][10]---------------*-[11][12][13][14]------------*-[15]-------------*-[16]-------------*-[17][18]-------------*-[19+]- Variable -----*
	 *  | uint8_t flags | uint16_t objectType | uint32_t objectInstance | uint32_t propertyIdentifier | uint32_t propertyArrayIndex | uint8_t dataType | uint8_t priority | uint16_t valueLength | char * valueAsString |
	 *  *---------------*---------------------*-------------------------*-----------------------------*-----------------------------*------------------*------------------*----------------------*----------------------*
	 */
	uint8_t writePropertyBuffer[100];
	memset(writePropertyBuffer, 0, 100);

	uint8_t offset = 0;
	uint32_t objectInstance = 0;
	uint16_t valueLength = 0; 


	// 0 
	objectInstance = 1;
	valueLength = strlen(WriteValueOne);
	writePropertyBuffer[0 + offset] = 0x00; // Flags 0x01 == use propertyArrayIndex, 0x02 == use priority 
	memcpy(writePropertyBuffer + 1 + offset, &CASBACnetStackExampleConstants::OBJECT_TYPE_ANALOG_OUTPUT, sizeof(uint16_t)); // objectType
	memcpy(writePropertyBuffer + 3 + offset, &objectInstance, sizeof(uint32_t)); // objectInstance
	memcpy(writePropertyBuffer + 7 + offset, &CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, sizeof(uint32_t)); // propertyIdentifier
	writePropertyBuffer[15 + offset] = CASBACnetStackExampleConstants::BACNET_DATATYPE_REAL; // dataType 	
	memcpy(writePropertyBuffer + 17 + offset, &valueLength, sizeof(uint16_t)); // valueLength
	memcpy(writePropertyBuffer + 19 + offset, WriteValueOne, valueLength); // valueAsString
	offset += 19 + valueLength; 


	// 1 
	objectInstance = 4;
	valueLength = strlen(WriteValueTwo);
	writePropertyBuffer[0 + offset] = 0x00; // Flags 0x01 == use propertyArrayIndex, 0x02 == use priority 
	memcpy(writePropertyBuffer + 1 + offset, &CASBACnetStackExampleConstants::OBJECT_TYPE_BINARY_OUTPUT, sizeof(uint16_t)); // objectType
	memcpy(writePropertyBuffer + 3 + offset, &objectInstance, sizeof(uint32_t)); // objectInstance
	memcpy(writePropertyBuffer + 7 + offset, &CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, sizeof(uint32_t)); // propertyIdentifier
	writePropertyBuffer[15 + offset] = CASBACnetStackExampleConstants::BACNET_DATATYPE_ENUMERATED; // dataType 	
	memcpy(writePropertyBuffer + 17 + offset, &valueLength, sizeof(uint16_t)); // valueLength
	memcpy(writePropertyBuffer + 19 + offset , WriteValueTwo, valueLength); // valueAsString
	offset += 19 + valueLength;
	
	// 2 
	objectInstance = 14;
	valueLength = strlen(WriteValueThree); 
	uint32_t propertyArrayIndex = 5; 
	writePropertyBuffer[0 + offset] = 0x01 && 0x02; // Flags 0x01 == use propertyArrayIndex, 0x02 == use priority 
	memcpy(writePropertyBuffer + 1 + offset, &CASBACnetStackExampleConstants::OBJECT_TYPE_MULTI_STATE_OUTPUT, sizeof(uint16_t)); // objectType
	memcpy(writePropertyBuffer + 3 + offset, &objectInstance, sizeof(uint32_t)); // objectInstance
	memcpy(writePropertyBuffer + 7 + offset, &CASBACnetStackExampleConstants::PROPERTY_IDENTIFIER_PRESENT_VALUE, sizeof(uint32_t)); // propertyIdentifier
	memcpy(writePropertyBuffer + 11 + offset, &propertyArrayIndex, sizeof(uint32_t)); // propertyArrayIndex	
	writePropertyBuffer[15 + offset] = CASBACnetStackExampleConstants::BACNET_DATATYPE_ENUMERATED; // dataType 	
	writePropertyBuffer[16 + offset] = 8; // priority 	
	memcpy(writePropertyBuffer + 17 + offset, &valueLength, sizeof(uint16_t)); // valueLength
	memcpy(writePropertyBuffer + 19 + offset, WriteValueThree, valueLength); // valueAsString
	offset += 19 + valueLength;

	
	if (!fpSendWritePropertyAsync(&invokeId, connectionString, 6, 0, 0, NULL, 0, 3, (void *)writePropertyBuffer)) {
		CLogging::Error("Failed to send SendReadPropertyAsync");
		return;
	}

	WaitForResponse();
}



